#include <assert.h>
#include <algorithm>
#include <fstream>
#include <set>
#include <stdexcept>
#include <sstream>
#include <oiran/math.hpp>
#include "field.hpp"


namespace {


	class block_holder_t
	{
	public:
		block_t** block;
		double distance;

		block_holder_t(block_t** block, double distance) :
				block(block),
				distance(distance)
		{}

		bool operator<(const block_holder_t& rhs) const
		{ return distance < rhs.distance; }
	};


} // unnamed namespace.


field_t::field_t(int stage) :
		M_stage(stage),
		M_valid(false)
{
	std::fill(M_blocks, M_blocks + field_width * field_height,
			reinterpret_cast<block_t*>(0));
}

field_t::~field_t()
{
	clear();
}

void
field_t::load()
{
	clear();

	std::ostringstream oss;
	oss << "stage/" << M_stage << ".dat";

	std::ifstream in(oss.str().c_str());
	if(in)
	{
		char load_data[field_width * field_height];
		in.read(load_data, sizeof(load_data));

		for(int y = 0, i = 0; y < field_height; ++y)
			for(int x = 0; x < field_width; ++x, ++i)
				M_blocks[y * field_width + x] = create_block(
						load_data[i], vector_type(x * block_size, y * block_size));

		M_valid = true;
	}
	else
		M_valid = false;
}

void
field_t::save()
{
	if(!M_valid)
		return;

	char save_data[field_width * field_height];
	block_t** now = M_blocks;
	for(int i = 0; i < field_width * field_height; ++i, ++now)
		save_data[i] = *now ? (*now)->id() : 0;

	std::ostringstream oss;
	oss << "stage/" << M_stage << ".dat";

	std::ofstream out(oss.str().c_str());
	if(!out)
	{
		std::ostringstream message;
		message << "failed to save : stage/" << M_stage << ".dat";
		throw std::runtime_error(message.str());
	}

	out.write(save_data, sizeof(save_data));
}

void
field_t::clear()
{
	block_t** now = M_blocks;
	for(int i = 0; i < field_width * field_height; ++i, ++now)
		if(*now)
		{
			delete *now;
			*now = 0;
		}
}

void
field_t::put(char id, int x, int y)
{
	assert(0 <= x && x < field_width &&
			0 <= y && y < field_height && "out of field range.");

	block_t** now = M_blocks + y * field_width + x;
	if(*now)
	{
		if((*now)->id() == id)
			return;
		delete *now;
	}
	*now = create_block(id, vector_type(x * block_size, y * block_size));

	M_valid = true;
}

void
field_t::draw(oiran::linda::graphics_t& g)
{
	if(!M_valid)
		return;

	block_t** now = M_blocks;
	for(int i = 0; i < field_width * field_height; ++i, ++now)
		if(*now)
			(*now)->draw(g);
}

void
field_t::collision_detection(ball_t& ball)
{
	if(ball.o == ball.p)
		return;

	int min_x = static_cast<int>(std::min(ball.o.x, ball.p.x) / block_size) - 1;
	int min_y = static_cast<int>(std::min(ball.o.y, ball.p.y) / block_size) - 1;
	int max_x = static_cast<int>(std::max(ball.o.x, ball.p.x) / block_size) + 1;
	int max_y = static_cast<int>(std::max(ball.o.y, ball.p.y) / block_size) + 1;
	min_x = oiran::constrain(min_x, 0, field_width - 1);
	max_x = oiran::constrain(max_x, 0, field_width - 1);
	min_y = oiran::constrain(min_y, 0, field_height - 1);
	max_y = oiran::constrain(max_y, 0, field_height - 1);

	// pick up to check blocks.
	std::multiset<block_holder_t> check_blocks;
	for(int y = min_y; y <= max_y; ++y)
		for(int x = min_x; x <= max_x; ++x)
		{
			block_t** now = M_blocks + y * field_width + x;
			if(*now)
				check_blocks.insert(
						block_holder_t(now, norm(ball.o - (*now)->center())));
		}

	if(check_blocks.empty())
		return;

	for(std::multiset<block_holder_t>::iterator i = check_blocks.begin();
			i != check_blocks.end(); ++i)
	{
		bool recheck = false;
		if((*(i->block))->collision_detection(ball, recheck))
		{
			if(!(*(i->block))->valid())
			{
				delete *(i->block);
				*(i->block) = 0;
			}

			if(recheck)
				collision_detection(ball);

			break;
		}
	}
}

bool
field_t::is_game_clear() const
{
	for(int y = 0; y < field_height; ++y)
		for(int x = 0; x < field_width; ++x)
		{
			block_t* now = M_blocks[y * field_width + x];
			if(now && now->id() != wall)
				return false;
		}
	return true;
}
