#ifndef OIRAN_RITA_VECTOR_HPP
#define OIRAN_RITA_VECTOR_HPP

#include <cmath>
#include <cstddef>
#include <algorithm>
#include <functional>
#include <iterator>
#include <numeric>
#include <oiran/rita/vector_base.hpp>

namespace oiran { namespace rita {


	template <typename T, std::size_t N>
	struct vector_t : public detail::vector_base_t<T, N>::type
	{
		// type definitions.
		typedef typename detail::vector_base_t<T, N>::type base_type;
		typedef T value_type;
		typedef T* iterator;
		typedef const T* const_iterator;
		typedef T& reference;
		typedef const T& const_reference;
		typedef std::reverse_iterator<iterator> reverse_iterator;
		typedef std::reverse_iterator<const_iterator> const_reverse_iterator;
		typedef std::size_t size_type;
		typedef std::ptrdiff_t difference_type;

		// constructors.
		vector_t() {}
		vector_t(const T& a, const T& b) : base_type(a, b) {}
		vector_t(const T& a, const T& b, const T& c) : base_type(a, b, c) {}
		explicit vector_t(const T* a) : base_type(a) {}
		template <typename U>
		vector_t(const vector_t<U, N>& v) : base_type(v) {}

		// iterators.
		iterator begin() { return base_type::elems; }
		const_iterator begin() const { return base_type::elems; }
		iterator end() { return base_type::elems + N; }
		const_iterator end() const { return base_type::elems + N; }
		reverse_iterator rbegin() { return reverse_iterator(end()); }
		const_reverse_iterator rbegin() const
		{ return const_reverse_iterator(end()); }
		reverse_iterator rend() { return reverse_iterator(begin()); }
		const_reverse_iterator rend() const
		{ return const_reverse_iterator(begin()); }

		// operator[].
		reference operator[](size_type i)
		{ return base_type::elems[i]; }
		const_reference operator[](size_type i) const
		{ return base_type::elems[i]; }

		// front, back.
		reference front() { return base_type::elems[0]; }
		const_reference front() const { return base_type::elems[0]; }
		reference back() { return base_type::elems[N - 1]; }
		const_reference back() const { return base_type::elems[N - 1]; }

		// size.
		static size_type size() { return N; }
		static bool empty() { return false; }
		static size_type max_size() { return N; }
		enum { static_size = N };

		// swap.
		void swap(vector_t<T, N>& v)
		{ std::swap_ranges(begin(), end(), v.begin()); }

		// direct access.
		T* data() { return base_type::elems; }
		const T* data() const { return base_type::elems; }

		// operators.
		const vector_t& operator+=(const vector_t& v)
		{
			std::transform(begin(), end(), v.begin(), begin(), std::plus<T>());
			return *this;
		}

		const vector_t& operator-=(const vector_t& v)
		{
			std::transform(begin(), end(), v.begin(), begin(), std::minus<T>());
			return *this;
		}

		template <typename U>
		const vector_t& operator*=(const U& v)
		{
			base_type::mul(v);
			return *this;
		}

		template <typename U>
		const vector_t& operator/=(const U& v)
		{
			base_type::div(v);
			return *this;
		}

		vector_t operator-() const
		{ return vector_t(*this) *= -1; }
	};


	// comparisions.
	template <typename T, std::size_t N>
	inline bool
	operator==(const vector_t<T, N>& l, const vector_t<T, N>& r)
	{ return std::equal(l.begin(), l.end(), r.begin()); }

	template <typename T, std::size_t N>
	inline bool
	operator!=(const vector_t<T, N>& l, const vector_t<T, N>& r)
	{ return !(l == r); }


	// operators.
	template <typename T, std::size_t N>
	inline vector_t<T, N>
	operator+(const vector_t<T, N>& l, const vector_t<T, N>& r)
	{ return vector_t<T, N>(l) += r; }

	template <typename T, std::size_t N>
	inline vector_t<T, N>
	operator-(const vector_t<T, N>& l, const vector_t<T, N>& r)
	{ return vector_t<T, N>(l) -= r; }

	template <typename T, std::size_t N, typename U>
	inline vector_t<T, N>
	operator*(const vector_t<T, N>& l, const U& r)
	{ return vector_t<T, N>(l) *= r; }

	template <typename T, typename U, std::size_t N>
	inline vector_t<U, N>
	operator*(const T& l, const vector_t<U, N>& r)
	{ return vector_t<U, N>(r) *= l; }

	template <typename T, std::size_t N, typename U>
	inline vector_t<T, N>
	operator/(const vector_t<T, N>& l, const U& r)
	{ return vector_t<T, N>(l) /= r; }


	// dot.
	template <typename T, std::size_t N>
	inline T
	dot(const vector_t<T, N>& l, const vector_t<T, N>& r)
	{ return std::inner_product(l.begin(), l.end(), r.begin(), T(0)); }


	// cross.
	template <typename T>
	inline T
	cross(const vector_t<T, 2>& l, const vector_t<T, 2>& r)
	{ return l.x * r.y - l.y * r.x; }

	template <typename T>
	inline vector_t<T, 3>
	cross(const vector_t<T, 3>& l, const vector_t<T, 3>& r)
	{
		return vector_t<T, 3>(
				l.y * r.z - l.z * r.y,
				l.z * r.x - l.x * r.z,
				l.x * r.y - l.y * r.x);
	}


	// norm2, norm, normalize.
	template <typename T, std::size_t N>
	inline T
	norm2(const vector_t<T, N>& v)
	{ return dot(v, v); }

	template <typename T, std::size_t N>
	inline T
	norm(const vector_t<T, N>& v)
	{ return std::sqrt(norm2(v)); }

	template <typename T, std::size_t N>
	inline vector_t<T, N>
	normalize(const vector_t<T, N>& v)
	{ return v / norm(v); }


	// proj.
	template <typename T, std::size_t N>
	inline vector_t<T, N>
	proj(const vector_t<T, N>& a, const vector_t<T, N>& b)
	{ return dot(a, b) / norm2(b) * b; }


	// perp.
	template <typename T, std::size_t N>
	inline vector_t<T, N>
	perp(const vector_t<T, N>& a, const vector_t<T, N>& b)
	{ return a - proj(a, b); }

	
}} // namespace oiran::rita.

#endif // OIRAN_RITA_VECTOR_HPP
